from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute, AddressableObject, AddressableDict
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class LightsStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.lights = AddressableDict(localAddress='lights', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update light status from dict')

        if 'value' in fromDict:
            if 'lights' in fromDict['value'] and fromDict['value']['lights'] is not None:
                for lightDict in fromDict['value']['lights']:
                    if 'name' in lightDict:
                        if lightDict['name'] in self.lights:
                            self.lights[lightDict['name']].update(fromDict=lightDict)
                        else:
                            self.lights[lightDict['name']] = LightsStatus.Light(fromDict=lightDict, parent=self.lights)
                for lightName in [lightName for lightName in self.lights.keys()
                                  if lightName not in [light['name'] for light in fromDict['value']['lights'] if 'name' in light]]:
                    del self.lights[lightName]
            else:
                self.lights.clear()
                self.lights.enabled = False
        else:
            self.lights.clear()
            self.lights.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['lights']))

    def __str__(self):
        string = super().__str__()
        if len(self.lights) > 0:
            string += f'\n\tLights: {len(self.lights)} items'
            for light in self.lights.values():
                string += f'\n\t\t{light}'
        return string

    class Light(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.status = AddressableAttribute(localAddress='status', parent=self,
                                               value=None, valueType=LightsStatus.Light.LightState)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update light from dict')

            if 'name' in fromDict:
                self.id = fromDict['name']
                self.localAddress = self.id
            else:
                LOG.error('Light is missing name attribute')

            self.status.fromDict(fromDict, 'status')

            for key, value in {key: value for key, value in fromDict.items() if key not in ['name', 'status']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            return f'{self.id}: {self.status.value.value}'  # pylint: disable=no-member

        class LightState(Enum,):
            ON = 'on'
            OFF = 'off'
            UNKNOWN = 'unknown open state'



from weconnect.auth.openid_session import OpenIDSession


class VWWebSession(OpenIDSession):
    def __init__(self, sessionuser, **kwargs):
        super(VWWebSession, self).__init__(**kwargs)
        self.sessionuser = sessionuser



from typing import Dict, Optional, Match

import re
import json
import logging
import requests

from urllib.parse import parse_qsl, urlsplit

from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from oauthlib.common import add_params_to_uri
from oauthlib.oauth2 import InsecureTransportError, is_secure_transport


from requests.models import CaseInsensitiveDict


from weconnect.auth.openid_session import AccessType
from weconnect.auth.vw_web_session import VWWebSession
from weconnect.errors import APICompatibilityError, AuthentificationError, RetrievalError, TemporaryAuthentificationError

LOG = logging.getLogger("weconnect")


class WeChargeSession(VWWebSession):
    def __init__(self, sessionuser, **kwargs):
        super(WeChargeSession, self).__init__(client_id='0fa5ae01-ebc0-4901-a2aa-4dd60572ea0e@apps_vw-dilab_com',
                                              refresh_url='https://identity.vwgroup.io/oidc/v1/token',
                                              scope='openid profile address email cars vin',
                                              redirect_uri='wecharge://authenticated',
                                              state=None,
                                              sessionuser=sessionuser,
                                              **kwargs)

        self.headers = CaseInsensitiveDict({
            'accept': '*/*',
            'content-type': 'application/json',
            'content-version': '1',
            'x-newrelic-id': 'VgAEWV9QDRAEXFlRAAYPUA==',
            'user-agent': 'WeConnect/3 CFNetwork/1327.0.4 Darwin/21.2.0',
            'accept-language': 'de-de',
        })

    @property
    def wcAccessToken(self):
        if self._token is not None and 'wc_access_token' in self._token:
            return self._token.get('wc_access_token')
        return None

    def login(self):
        super(WeChargeSession, self).login()
        authorizationUrl = self.authorizationUrl(url='https://identity.vwgroup.io/oidc/v1/authorize')
        response = self.doWebAuth(authorizationUrl)
        self.fetchTokens('https://wecharge.apps.emea.vwapps.io/user-identity/v1/identity/login',
                         authorization_response=response)

    def refresh(self):
        self.refreshTokens(
            'https://wecharge.apps.emea.vwapps.io/user-identity/v1/identity/login',
        )

    def doWebAuth(self, authorizationUrl):  # noqa: C901
        websession: requests.Session = requests.Session()
        retries = Retry(total=self.retries,
                        backoff_factor=0.1,
                        status_forcelist=[500],
                        raise_on_status=False)
        websession.mount('https://', HTTPAdapter(max_retries=retries))
        websession.headers = CaseInsensitiveDict({
            'user-agent': 'Mozilla/5.0 (Linux; Android 10) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 '
                          'Chrome/74.0.3729.185 Mobile Safari/537.36',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,'
                      'application/signed-exchange;v=b3',
            'accept-language': 'en-US,en;q=0.9',
            'accept-encoding': 'gzip, deflate',
            'x-requested-with': 'de.volkswagen.carnet.eu.eremote',
            'upgrade-insecure-requests': '1',
        })

        while True:
            loginFormResponse: requests.Response = websession.get(authorizationUrl, allow_redirects=False)
            if loginFormResponse.status_code == requests.codes['ok']:
                break
            elif loginFormResponse.status_code == requests.codes['found']:
                if 'Location' in loginFormResponse.headers:
                    authorizationUrl = loginFormResponse.headers['Location']
                else:
                    raise APICompatibilityError('Forwarding without Location in Header')
            elif loginFormResponse.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during login')
            else:
                raise APICompatibilityError('Retrieving credentials page was not successfull,'
                                            f' status code: {loginFormResponse.status_code}')
        # Find login form on page to obtain inputs
        emailFormRegex = r'<form.+id=\"emailPasswordForm\".*action=\"(?P<formAction>[^\"]+)\"[^>]*>' \
            r'(?P<formContent>.+?(?=</form>))</form>'
        match: Optional[Match[str]] = re.search(emailFormRegex, loginFormResponse.text, flags=re.DOTALL)
        if match is None:
            raise AuthentificationError('No login email form found')
        # retrieve target url from form
        target: str = match.groupdict()['formAction']

        # Find all inputs and put those in formData dictionary
        inputRegex = r'<input[\\n\\r\s][^/]*name=\"(?P<name>[^\"]+)\"([\\n\\r\s]value=\"(?P<value>[^\"]+)\")?[^/]*/>'
        formData: Dict[str, str] = {}
        for match in re.finditer(inputRegex, match.groupdict()['formContent']):
            if match.groupdict()['name']:
                formData[match.groupdict()['name']] = match.groupdict()['value']
        if not all(x in ['_csrf', 'registerFlow', 'relayState', 'hmac', 'identifier'] for x in formData):
            raise AuthentificationError('Could not find all required input fields in login page')

        # Set email to the provided username
        formData['identifier'] = self.sessionuser.username

        # build url from form action
        login2Url: str = 'https://identity.vwgroup.io' + target

        loginHeadersForm: CaseInsensitiveDict = websession.headers.copy()
        loginHeadersForm['Content-Type'] = 'application/x-www-form-urlencoded'

        # Post form content and retrieve credentials page
        login2Response: requests.Response = websession.post(login2Url, headers=loginHeadersForm, data=formData, allow_redirects=True)

        if login2Response.status_code != requests.codes['ok']:  # pylint: disable=E1101
            if login2Response.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during login')
            raise AuthentificationError('Retrieving credentials page was not successfull,'
                                        f' status code: {login2Response.status_code}')

        credentialsTemplateRegex = r'<script>\s+window\._IDK\s+=\s+\{\s' \
            r'(?P<templateModel>.+?(?=\s+\};?\s+</script>))\s+\};?\s+</script>'
        match = re.search(credentialsTemplateRegex, login2Response.text, flags=re.DOTALL)
        if match is None:
            raise AuthentificationError('No credentials form found')
        if match.groupdict()['templateModel']:
            lineRegex = r'\s*(?P<name>[^\:]+)\:\s+[\'\{]?(?P<value>.+)[\'\}][,]?'
            form2Data: Dict[str, str] = {}
            for match in re.finditer(lineRegex, match.groupdict()['templateModel']):
                if match.groupdict()['name'] == 'templateModel':
                    templateModelString = '{' + match.groupdict()['value'] + '}'
                    if templateModelString.endswith(','):
                        templateModelString = templateModelString[:-len(',')]
                    templateModel = json.loads(templateModelString)
                    if 'relayState' in templateModel:
                        form2Data['relayState'] = templateModel['relayState']
                    if 'hmac' in templateModel:
                        form2Data['hmac'] = templateModel['hmac']
                    if 'emailPasswordForm' in templateModel and 'email' in templateModel['emailPasswordForm']:
                        form2Data['email'] = templateModel['emailPasswordForm']['email']
                    if 'error' in templateModel and templateModel['error'] is not None:
                        if templateModel['error'] == 'validator.email.invalid':
                            raise AuthentificationError('Error during login, email invalid')
                        raise AuthentificationError(f'Error during login: {templateModel["error"]}')
                    if 'registerCredentialsPath' in templateModel and templateModel['registerCredentialsPath'] == 'register':
                        raise AuthentificationError(f'Error during login, account {self.sessionuser.username} does not exist')
                    if 'errorCode' in templateModel:
                        raise AuthentificationError('Error during login, is the username correct?')
                    if 'postAction' in templateModel:
                        target = templateModel['postAction']
                    else:
                        raise APICompatibilityError('Form does not contain postAction')
                elif match.groupdict()['name'] == 'csrf_token':
                    form2Data['_csrf'] = match.groupdict()['value']
        form2Data['password'] = self.sessionuser.password
        if not all(x in ['_csrf', 'relayState', 'hmac', 'email', 'password'] for x in form2Data):
            raise AuthentificationError('Could not find all required input fields in credentials page')

        login3Url = f'https://identity.vwgroup.io/signin-service/v1/{self.client_id}/{target}'

        # Post form content and retrieve userId in forwarding Location
        login3Response: requests.Response = websession.post(login3Url, headers=loginHeadersForm, data=form2Data, allow_redirects=False)
        if login3Response.status_code not in (requests.codes['found'], requests.codes['see_other']):
            if login3Response.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during login')
            raise AuthentificationError('Forwarding expected (status code 302),'
                                        f' but got status code {login3Response.status_code}')
        if 'Location' not in login3Response.headers:
            raise AuthentificationError('No url for forwarding in response headers')

        # Parse parametes from forwarding url
        params: Dict[str, str] = dict(parse_qsl(urlsplit(login3Response.headers['Location']).query))

        # Check if error
        if 'error' in params and params['error']:
            errorMessages: Dict[str, str] = {
                'login.errors.password_invalid': 'Password is invalid',
                'login.error.throttled': 'Login throttled, probably too many wrong logins. You have to wait some'
                                         ' minutes until a new login attempt is possible'
            }
            if params['error'] in errorMessages:
                error = errorMessages[params['error']]
            else:
                error = params['error']
            raise AuthentificationError(error)

        # Check for user id
        if 'userId' not in params or not params['userId']:
            if 'updated' in params and params['updated'] == 'dataprivacy':
                raise AuthentificationError('You have to login at myvolkswagen.de and accept the terms and conditions')
            raise AuthentificationError('No user id provided')
        self.__userId = params['userId']  # pylint: disable=unused-private-member

        # Now follow the forwarding until forwarding URL starts with 'weconnect://authenticated#'
        afterLoginUrl: str = login3Response.headers['Location']

        while True:
            if 'consent' in afterLoginUrl:
                consentURL = afterLoginUrl
            afterLoginResponse = self.get(afterLoginUrl, allow_redirects=False, access_type=AccessType.NONE)
            if afterLoginResponse.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during login')

            if 'Location' not in afterLoginResponse.headers:
                if consentURL is not None:
                    raise AuthentificationError('It seems like you need to accept the terms and conditions for the WeConnect ID service.'
                                                f' Try to visit the URL "{consentURL}" or log into the WeConnect ID smartphone app')
                raise AuthentificationError('No Location for forwarding in response headers')

            afterLoginUrl = afterLoginResponse.headers['Location']

            if afterLoginUrl.startswith(self.redirect_uri):
                break

        if afterLoginUrl.startswith(self.redirect_uri + '#'):
            queryurl = afterLoginUrl.replace(self.redirect_uri + '#', 'https://egal?')
        else:
            queryurl = afterLoginUrl
        return queryurl

    def fetchTokens(
        self,
        token_url,
        authorization_response=None,
        **kwargs
    ):
        self.parseFromFragment(authorization_response)

        if all(key in self.token for key in ('state', 'id_token', 'access_token', 'code')):
            loginHeadersForm: CaseInsensitiveDict = self.headers.copy()
            loginHeadersForm['accept'] = 'application/json'
            loginHeadersForm["x-api-key"] = "yabajourasW9N8sm+9F/oP=="

            urlParams = [(('redirect_uri', self.redirect_uri)),
                         (('code', self.token["code"]))]
            token_url = add_params_to_uri(token_url, urlParams)

            tokenResponse = self.get(token_url, headers=loginHeadersForm, allow_redirects=False, access_type=AccessType.ID)
            if tokenResponse.status_code != requests.codes['ok']:
                raise TemporaryAuthentificationError(f'Token could not be fetched due to temporary WeConnect failure: {tokenResponse.status_code}')

            self.parseFromBody(tokenResponse.text)

            return self.token

    def refreshTokens(
        self,
        token_url,
        refresh_token=None,
        auth=None,
        timeout=None,
        headers=None,
        verify=True,
        proxies=None,
        **kwargs
    ):
        LOG.info('Refreshing tokens')
        if not token_url:
            raise ValueError("No token endpoint set for auto_refresh.")

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.refresh_token
        if refresh_token is None:
            raise ValueError("Missing refresh token.")

        if headers is None:
            headers = self.headers

        urlParams = [(('redirect_uri', self.redirect_uri)),
                     (('refresh_token', refresh_token))]

        token_url = add_params_to_uri(token_url, urlParams)

        refreshHeaders: CaseInsensitiveDict = headers.copy()
        refreshHeaders['accept'] = 'application/json'
        refreshHeaders["x-api-key"] = "yabajourasW9N8sm+9F/oP=="

        tokenResponse = self.get(
            token_url,
            auth=auth,
            timeout=timeout,
            headers=refreshHeaders,
            verify=verify,
            withhold_token=False,
            proxies=proxies,
            access_type=AccessType.REFRESH
        )
        if tokenResponse.status_code == requests.codes['unauthorized']:
            raise AuthentificationError('Refreshing tokens failed: Server requests new authorization')
        elif tokenResponse.status_code in (requests.codes['internal_server_error'], requests.codes['service_unavailable'], requests.codes['gateway_timeout']):
            raise TemporaryAuthentificationError('Token could not be refreshed due to temporary WeConnect failure: {tokenResponse.status_code}')
        elif tokenResponse.status_code == requests.codes['ok']:
            self.parseFromBody(tokenResponse.text)
            if "refresh_token" not in self.token:
                LOG.debug("No new refresh token given. Re-using old.")
                self.token["refresh_token"] = refresh_token
            return self.token
        else:
            raise RetrievalError(f'Status Code from WeConnect while refreshing tokens was: {tokenResponse.status_code}')

    def addToken(self, uri, body=None, headers=None, access_type=AccessType.ACCESS, **kwargs):
        headers = headers or {}
        uri, headers, body = super(WeChargeSession, self).addToken(uri, body=body, headers=headers, access_type=access_type, **kwargs)

        if access_type == AccessType.ACCESS:
            if not (self.wcAccessToken):
                raise ValueError("Missing wc access token.")
            headers['wc_access_token'] = self.wcAccessToken

        return (uri, headers, body)



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class GenericRequestStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.status = AddressableAttribute(localAddress='status', parent=self,
                                           value=None, valueType=GenericRequestStatus.Status)
        self.group = AddressableAttribute(localAddress='group', parent=self, value=None, valueType=int)
        self.info = AddressableAttribute(localAddress='info', parent=self, value=None, valueType=str)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Request status from dict')

        self.status.fromDict(fromDict, 'status')
        self.group.fromDict(fromDict, 'group')
        self.info.fromDict(fromDict, 'info')

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['status', 'group', 'info']))

    def __str__(self):
        string = super().__str__() + '\n'
        if self.status.enabled:
            string += f'\tStatus: {self.status.value.value}\n'  # pylint: disable=no-member
        if self.group.enabled:
            string += f'\tGroup: {self.group.value}\n'
        if self.info.enabled:
            string += f'\tInfo: {self.info.value}\n'
        return string

    class Status(Enum,):
        SUCCESSFULL = 'successful'
        FAIL = 'fail'
        POLLING_TIMEOUT = 'polling_timeout'
        IN_PROGRESS = 'in_progress'
        QUEUED = 'queued'
        DELAYED = 'delayed'
        TIMEOUT = 'timeout'
        FAIL_VEHICLE_IS_OFFLINE = 'fail_vehicle_is_offline'
        FAIL_IGNITION_ON = 'fail_ignition_on'
        FAIL_BATTERY_LOW = 'fail_battery_low'
        UNKNOWN = 'unknown status'



from enum import Enum
from typing import List
import logging

from weconnect.addressable import AddressableAttribute, AddressableObject
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class ChargeMode(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.preferredChargeMode = AddressableAttribute(localAddress='preferredChargeMode', value=None, parent=self,
                                                        valueType=ChargeMode.ChargeModeEnum)
        self.availableChargeModes = ChargeMode.ChargeModeList(localAddress='availableChargeModes', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update ChargeMode status from dict')

        if 'value' in fromDict:
            self.preferredChargeMode.fromDict(fromDict['value'], 'preferredChargeMode')

            if 'availableChargeModes' in fromDict['value']:
                if self.availableChargeModes is None:
                    self.availableChargeModes = ChargeMode.ChargeModeList(localAddress='availableChargeModes', parent=self,
                                                                          fromDict=fromDict['value']['availableChargeModes'])
                else:
                    self.availableChargeModes.update(fromDict=fromDict['value']['availableChargeModes'])
            elif self.availableChargeModes is not None:
                self.availableChargeModes.clear()
                self.availableChargeModes.enabled = False
        else:
            self.preferredChargeMode.enabled = False
            self.availableChargeModes.clear()
            self.availableChargeModes.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['preferredChargeMode', 'availableChargeModes']))

    def __str__(self):
        string = super().__str__()
        if self.preferredChargeMode is not None and self.preferredChargeMode.enabled:
            string += f'\n\tPreferred charge mode: {self.preferredChargeMode.value.value}'  # pylint: disable=no-member
        if self.availableChargeModes is not None and self.availableChargeModes.enabled:
            string += f'\n\tAvailable charge modes: {self.availableChargeModes}'
        return string

    class ChargeModeEnum(Enum,):
        MANUAL = 'manual'
        TIMER = 'timer'
        ONLY_OWN_CURRENT = 'onlyOwnCurrent'
        PREFERRED_CHARGING_TIMES = 'preferredChargingTimes'
        TIMER_CHARGING_WITH_CLIMATISATION = 'timerChargingWithClimatisation'
        INVALID = 'invalid'
        UNKNOWN = 'unknown charge mode'

    class ChargeModeList(AddressableObject, List):
        def update(self, fromDict):
            LOG.debug('Update timer from dict')

            self.clear()
            if fromDict is not None and len(fromDict) > 0:
                for mode in fromDict:
                    try:
                        self.append(ChargeMode.ChargeModeEnum(mode))
                    except ValueError:
                        self.append(ChargeMode.ChargeModeEnum.UNKNOWN)
                        LOG.warning('An unsupported mode: %s was provided, please report this as a bug', mode)
                if not self.enabled:
                    self.enabled = True
            elif self.enabled:
                self.enabled = False

        def __str__(self):
            return '[' + ', '.join([item.value for item in self]) + ']'



from __future__ import annotations
from typing import Callable, NoReturn, Optional, Dict, List, Set, Any, Tuple, Union, Type, TypeVar, Generic

import json
import logging
import time as timemodule
from datetime import datetime, timezone, time
from enum import Enum, IntEnum, Flag, auto

from weconnect.util import toBool, robustTimeParse, ExtendedWithNullEncoder

SUPPORT_IMAGES = False
try:
    from PIL import Image  # type: ignore
    from weconnect.util import imgToASCIIArt  # pylint: disable=ungrouped-imports
    SUPPORT_IMAGES = True
except ImportError:
    pass

SUPPORT_ASCII_IMAGES = False
try:
    import ascii_magic  # type: ignore
    SUPPORT_ASCII_IMAGES = True
except ImportError:
    pass

LOG: logging.Logger = logging.getLogger("weconnect")


class AddressableLeaf():
    def __init__(
        self,
        localAddress: str,
        parent: Optional[AddressableObject],
    ) -> None:
        self.__enabled: bool = False
        self.__localAddress: str = localAddress
        self.__parent: Optional[AddressableObject] = parent
        self.__observers: Set[Tuple[Callable[[Optional[Any], AddressableLeaf.ObserverEvent], None],
                                    AddressableLeaf.ObserverEvent, AddressableLeaf.ObserverPriority, bool]] = set()
        self.lastChange: Optional[datetime] = None
        self.lastUpdateFromServer: Optional[datetime] = None
        self.lastUpdateFromCar: Optional[datetime] = None
        self.onCompleteNotifyFlags: Optional[AddressableLeaf.ObserverEvent] = None

    def __del__(self) -> None:
        if self.enabled:
            self.enabled = False

    def addObserver(self, observer: Callable, flag: AddressableLeaf.ObserverEvent, priority: Optional[AddressableLeaf.ObserverPriority] = None,
                    onUpdateComplete: bool = False) -> None:
        if priority is None:
            priority = AddressableLeaf.ObserverPriority.USER_MID
        self.__observers.add((observer, flag, priority, onUpdateComplete))
        LOG.debug('%s: Observer added with flags: %s', self.getGlobalAddress(), flag)

    def removeObserver(self, observer: Callable, flag: Optional[AddressableLeaf.ObserverEvent] = None) -> None:
        self.__observers = filter(lambda observerEntry: observerEntry[0] == observer
                                  or (flag is not None and observerEntry[1] == flag), self.__observers)

    def getObservers(self, flags, onUpdateComplete: bool = False) -> List[Any]:
        return [observerEntry[0] for observerEntry in self.getObserverEntries(flags, onUpdateComplete)]

    def getObserverEntries(self, flags: AddressableLeaf.ObserverEvent, onUpdateComplete: bool = False) -> List[Any]:
        observers: Set[Tuple[Callable, AddressableLeaf.ObserverEvent, AddressableLeaf.ObserverPriority, bool]] = set()
        for observerEntry in self.__observers:
            observer, observerflags, priority, observerOnUpdateComplete = observerEntry
            del observer
            del priority
            if (flags & observerflags) and observerOnUpdateComplete == onUpdateComplete:
                observers.add(observerEntry)
        if self.__parent is not None:
            observers.update(self.__parent.getObserverEntries(flags, onUpdateComplete))
        return sorted(observers, key=lambda entry: int(entry[2]))

    def notify(self, flags: AddressableLeaf.ObserverEvent) -> None:
        observers: List[Callable] = self.getObservers(flags, onUpdateComplete=False)
        for observer in observers:
            observer(element=self, flags=flags)
        if self.onCompleteNotifyFlags is not None:
            # Remove disabled if was enabled and not yet notified
            if (flags & AddressableLeaf.ObserverEvent.ENABLED) and (self.onCompleteNotifyFlags & AddressableLeaf.ObserverEvent.DISABLED):
                self.onCompleteNotifyFlags &= ~AddressableLeaf.ObserverEvent.DISABLED  # pylint: disable=invalid-unary-operand-type
            # Remove enabled if was enabled and not yet notified
            elif (flags & AddressableLeaf.ObserverEvent.DISABLED) and (self.onCompleteNotifyFlags & AddressableLeaf.ObserverEvent.ENABLED):
                self.onCompleteNotifyFlags &= ~AddressableLeaf.ObserverEvent.ENABLED  # pylint: disable=invalid-unary-operand-type
            else:
                self.onCompleteNotifyFlags |= flags
        else:
            self.onCompleteNotifyFlags = flags
        LOG.debug('%s: Notify called with flags: %s for %d observers', self.getGlobalAddress(), flags, len(observers))

    def updateComplete(self) -> None:
        if self.onCompleteNotifyFlags is not None:
            observers = self.getObservers(self.onCompleteNotifyFlags, onUpdateComplete=True)
            for observer in observers:
                observer(element=self, flags=self.onCompleteNotifyFlags)
            if len(observers) > 0:
                LOG.debug('%s: Notify called on update complete with flags: %s for %d observers', self.getGlobalAddress(),
                          self.onCompleteNotifyFlags, len(observers))
            self.onCompleteNotifyFlags = None

    @property
    def enabled(self) -> bool:
        return self.__enabled

    @enabled.setter
    def enabled(self, setEnabled: bool) -> None:
        if setEnabled and not self.__enabled:
            if self.parent is not None:
                self.parent.addChild(self)
            self.notify(AddressableLeaf.ObserverEvent.ENABLED)
        elif not setEnabled and self.__enabled:
            self.notify(AddressableLeaf.ObserverEvent.DISABLED)
        self.__enabled = setEnabled

    @property
    def localAddress(self) -> str:
        return self.getLocalAddress()

    @localAddress.setter
    def localAddress(self, newAdress: str) -> None:
        self.__localAddress = newAdress

    @property
    def parent(self) -> Optional[AddressableObject]:
        return self.__parent

    @parent.setter
    def parent(self, newParent: AddressableObject):
        self.__parent = newParent

    def getLocalAddress(self) -> str:
        return self.__localAddress

    def getGlobalAddress(self) -> str:
        address = ''
        if self.__parent is not None:
            address = f'{self.__parent.getGlobalAddress()}/'
        address += f'{self.__localAddress}'
        return address

    def getByAddressString(self, addressString: str) -> Union[AddressableLeaf, bool]:
        if addressString == self.getLocalAddress():
            return self
        if addressString == '..':
            if self.parent is None:
                return self
            return self.parent
        if addressString == '/':
            return self.getRoot()
        return False

    def getRoot(self) -> AddressableLeaf:
        if self.parent is None:
            return self
        return self.parent.getRoot()

    class ObserverEvent(Flag):
        ENABLED = auto()
        DISABLED = auto()
        VALUE_CHANGED = auto()
        UPDATED_FROM_SERVER = auto()
        UPDATED_FROM_CAR = auto()
        ALL = ENABLED | DISABLED | VALUE_CHANGED | UPDATED_FROM_SERVER | UPDATED_FROM_CAR

    class ObserverPriority(IntEnum):
        INTERNAL_FIRST = 1
        INTERNAL_HIGH = 2
        USER_HIGH = 3
        INTERNAL_MID = 4
        USER_MID = 5
        INTERNAL_LOW = 6
        USER_LOW = 7
        INTERNAL_LAST = 8


T = TypeVar('T')


class AddressableAttribute(AddressableLeaf, Generic[T]):
    def __init__(
        self,
        localAddress: str,
        parent: AddressableObject,
        value: Optional[T],
        valueType: Type[T],
        lastUpdateFromCar: Optional[datetime] = None,
        valueGetter: Optional[Callable[[], Optional[T]]] = None,
        valueSetter: Optional[Callable[[Optional[T]], None]] = None
    ) -> None:
        super().__init__(localAddress, parent)
        self.__value: Optional[T] = None
        self.valueType: Type[T] = valueType
        self.valueGetter = valueGetter
        self.valueSetter = valueSetter
        if value is not None:
            self.setValueWithCarTime(value, lastUpdateFromCar, fromServer=True)

    @property
    def value(self) -> Optional[T]:
        if self.valueGetter is not None:
            return self.valueGetter()
        return self.__value

    @value.setter
    def value(self, newValue: Optional[T]) -> NoReturn:
        raise NotImplementedError('You cannot set this attribute. Set is not implemented')

    def asDict(self, filterCallable: Optional[Callable[[Any], None]] = None):
        if filterCallable is None or not filterCallable(self.value):
            return self.value
        return None

    def toJSON(self):
        if SUPPORT_IMAGES and isinstance(self.value, Image.Image):
            return None
        return json.dumps(self.value, cls=ExtendedWithNullEncoder, skipkeys=True, indent=4)

    def setValueWithCarTime(self, newValue, lastUpdateFromCar: Optional[datetime] = None, fromServer: bool = False, noNotify: bool = False) -> None:
        if newValue is not None and not isinstance(newValue, self.valueType):
            raise ValueError(f'{self.getGlobalAddress()}: new value {newValue} must be of type {self.valueType}'
                             f' but is of type {type(newValue)}')
        valueChanged: bool = newValue != self.__value
        self.__value = newValue
        flags: Optional[AddressableLeaf.ObserverEvent] = None
        if not self.enabled and valueChanged:
            self.enabled = True
        self.lastUpdateFromServer = datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
        if valueChanged:
            self.lastChange = datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
            flags = AddressableLeaf.ObserverEvent.VALUE_CHANGED
            if fromServer:
                if flags is None:
                    flags = AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER
                else:
                    flags |= AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER
        elif fromServer:
            flags = AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER

        if lastUpdateFromCar is not None and lastUpdateFromCar != self.lastUpdateFromCar:
            self.lastUpdateFromCar = lastUpdateFromCar
            if flags is None:
                flags = AddressableLeaf.ObserverEvent.UPDATED_FROM_CAR
            else:
                flags |= AddressableLeaf.ObserverEvent.UPDATED_FROM_CAR

        if not noNotify and flags is not None:
            self.notify(flags)

    def isLeaf(self) -> bool:  # pylint: disable=R0201
        return True

    def saveToFile(self, filename: str) -> None:  # noqa: C901
        if self.value is not None:
            if filename.endswith(('.txt', '.TXT', '.text')):
                with open(filename, mode='w', encoding='utf8') as textfile:
                    if SUPPORT_IMAGES and SUPPORT_ASCII_IMAGES and isinstance(self.value, Image.Image):
                        textfile.write(imgToASCIIArt(self.value, columns=120, mode=ascii_magic.Modes.ASCII))
                    else:
                        textfile.write(str(self))
            elif filename.endswith(('.htm', '.HTM', '.html', '.HTML')):
                with open(filename, mode='w', encoding='utf8') as htmlfile:
                    if SUPPORT_IMAGES and SUPPORT_ASCII_IMAGES and isinstance(self.value, Image.Image):
                        html = """<!DOCTYPE html><head><title>ASCII art</title></head><body><pre style="display: inline-block; border-width: 4px 6px;
    border-color: black; border-style: solid; background-color:black; font-size: 8px;">"""
                        htmlfile.write(html)
                        htmlfile.write(imgToASCIIArt(self.value, columns=240, mode=ascii_magic.Modes.HTML))
                        htmlfile.write('<pre/></body></html>')
                    else:
                        htmlfile.write(str(self))
            elif filename.endswith(('.json')):
                with open(filename, mode='w', encoding='utf8') as textfile:
                    if SUPPORT_IMAGES and SUPPORT_ASCII_IMAGES and isinstance(self.value, Image.Image):
                        raise ValueError('Attribute is an image and cannot be converted to json')
                    textfile.write(self.toJSON() + '\n')
            elif filename.endswith(('.png', '.PNG')):
                with open(filename, mode='wb') as pngfile:
                    if SUPPORT_IMAGES and isinstance(self.value, Image.Image):
                        self.value.save(fp=pngfile, format='PNG')  # pylint: disable=no-member
                    else:
                        raise ValueError('Attribute is no image and cannot be converted to one')
            elif filename.endswith(('.jpg', '.JPG', '.jpeg', '.JPEG')):
                with open(filename, mode='wb') as jpgfile:
                    if SUPPORT_IMAGES and isinstance(self.value, Image.Image):
                        if self.value.mode in ("RGBA", "P"):  # pylint: disable=no-member
                            raise ValueError('Image contains transparency and thus cannot be saved as jpeg-file')
                        self.value.save(fp=jpgfile, format='JPEG')  # pylint: disable=no-member
                    else:
                        raise ValueError('Attribute is no image and cannot be converted to one')
            else:
                raise ValueError('I cannot recognize the target file extension')
        else:
            raise ValueError('I cannot save None value')

    def fromDict(self, fromDict: dict, key: str):  # noqa: C901
        if fromDict is not None and key in fromDict and fromDict[key] is not None:
            if issubclass(self.valueType, bool):
                self.setValueWithCarTime(toBool(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
            elif issubclass(self.valueType, int):
                self.setValueWithCarTime(int(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
            elif issubclass(self.valueType, float):
                self.setValueWithCarTime(float(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
            elif issubclass(self.valueType, Enum):
                if fromDict[key]:
                    try:
                        self.setValueWithCarTime(self.valueType(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
                    except ValueError:
                        self.setValueWithCarTime(self.valueType.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                        LOG.warning('%s: An unsupported %s: %s was provided, known values are [%s]'
                                    ' please report this as a bug', self.getGlobalAddress(), key, fromDict[key],
                                    ', '.join([state.value for state in list(self.valueType)]))
                else:
                    self.enabled = False
            elif issubclass(self.valueType, datetime):
                self.setValueWithCarTime(robustTimeParse(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
            elif issubclass(self.valueType, time):
                parsedtime = timemodule.strptime(fromDict[key], "%H:%M")
                self.setValueWithCarTime(time(hour=parsedtime.tm_hour, minute=parsedtime.tm_min), lastUpdateFromCar=None, fromServer=True)
            elif issubclass(self.valueType, str):
                self.setValueWithCarTime(str(fromDict[key]), lastUpdateFromCar=None, fromServer=True)
            else:
                raise ValueError(f'Unknown attribute type {self.valueType}')
        else:
            self.enabled = False

    def __str__(self) -> str:
        if isinstance(self.value, Enum):
            return str(self.value.value)  # pylint: disable=no-member
        if isinstance(self.value, datetime):
            return self.value.isoformat()  # pylint: disable=no-member
        if SUPPORT_IMAGES and isinstance(self.value, Image.Image):
            return imgToASCIIArt(self.value)  # pylint: disable=no-member
        return str(self.value)


class ChangeableAttribute(AddressableAttribute):
    def __init__(
        self,
        localAddress: str,
        parent: AddressableObject,
        value,
        valueType=str,
        lastUpdateFromCar: Optional[datetime] = None,
        valueGetter: Optional[Callable[[], Optional[T]]] = None,
        valueSetter: Optional[Callable[[Optional[T]], None]] = None
    ) -> None:
        super().__init__(localAddress=localAddress, parent=parent, value=value, valueType=valueType,
                         lastUpdateFromCar=lastUpdateFromCar, valueGetter=valueGetter, valueSetter=valueSetter)

    @AddressableAttribute.value.setter  # type: ignore # noqa: C901
    def value(self, newValue):  # noqa: C901
        exceptions = []

        iterablevalueType = self.valueType
        if not isinstance(iterablevalueType, tuple):
            iterablevalueType = (iterablevalueType,)
        for valType in iterablevalueType:  # pylint: disable=too-many-nested-blocks
            if isinstance(newValue, str) and valType != str:
                try:
                    if valType in [int, float]:
                        newValue = valType(newValue)
                        exceptions = []
                    elif valType == bool:
                        newValue = toBool(newValue)
                        exceptions = []
                    elif issubclass(valType, Enum):
                        if not isinstance(newValue, Enum):
                            newValue = valType(newValue)
                            exceptions = []
                        try:
                            allowedValues = valType.allowedValues()
                            if newValue not in allowedValues:
                                raise ValueError('Value is not in allowed values')
                        except AttributeError:
                            pass
                        newValue = valType(newValue)
                        exceptions = []
                        break
                except ValueError:
                    valueFormat = ''
                    if valType == int:
                        valueFormat = 'N (Decimal number)'
                    elif valType == float:
                        valueFormat = 'F.F (Floating Point Number)'
                    elif valType == bool:
                        valueFormat = 'True/False (Boolean)'
                    elif issubclass(valType, Enum):
                        try:
                            valueFormat = 'select one of [' + ', '.join([enum.value for enum in valType.allowedValues()]) + ']'
                        except AttributeError:
                            valueFormat = 'select one of [' + ', '.join([enum.value for enum in valType])
                    exceptions.append(valueFormat)
            elif isinstance(newValue, int) and valType != int:
                if valType == float:
                    newValue = float(newValue)
                    exceptions = []
                    break
            else:
                exceptions = []

        if exceptions:
            raise ValueError(f'id {self.getGlobalAddress()} cannot be set to value {newValue}.'
                             f' You need to provide it in the correct format {" or ".join(exceptions)}')
        if self.valueSetter is not None:
            self.valueSetter(newValue)
        else:
            self.setValueWithCarTime(newValue=newValue, lastUpdateFromCar=None)


class AliasChangeableAttribute(ChangeableAttribute):
    def __init__(
        self,
        localAddress: str,
        parent: AddressableObject,
        value,
        targetAttribute: ChangeableAttribute,
        conversion,
        valueType=str,
        lastUpdateFromCar: Optional[datetime] = None,
    ) -> None:
        super().__init__(localAddress=localAddress, parent=parent,
                         value=value, valueType=valueType,
                         lastUpdateFromCar=lastUpdateFromCar)
        self.targetAttribute = targetAttribute
        self.conversion = conversion

    @AddressableAttribute.value.setter
    def value(self, newValue):
        ChangeableAttribute.value.fset(self, newValue)  # pylint: disable=no-member
        convertedNewValue = self.conversion(self.value)
        self.targetAttribute.setValueWithCarTime(newValue=convertedNewValue,
                                                 lastUpdateFromCar=None)


class AddressableObject(AddressableLeaf):
    def __init__(
        self,
        localAddress: str,
        parent: Optional[AddressableObject],
    ) -> None:
        super().__init__(localAddress, parent)
        self.__children: dict[str, AddressableLeaf] = {}

    @AddressableLeaf.enabled.setter  # type: ignore
    def enabled(self, setEnabled: bool) -> None:
        if not setEnabled and self.enabled:
            for child in self.__children.values():
                child.enabled = False
        AddressableLeaf.enabled.fset(self, setEnabled)  # type: ignore

    def asDict(self, filterCallable: Optional[Callable[[Any], None]] = None):
        asDict = {}
        for child in self.children:
            if child.enabled:
                childDict = child.asDict(filterCallable=filterCallable)
                if childDict is not None:
                    asDict[child.getLocalAddress()] = childDict
        return asDict

    def toJSON(self):
        def filterDict(element):
            if SUPPORT_IMAGES and isinstance(element, Image.Image):
                return True
            return False
        return json.dumps(self.asDict(filterCallable=filterDict), cls=ExtendedWithNullEncoder, skipkeys=True, indent=4)

    def isLeaf(self) -> bool:
        return not self.__children

    def addChild(self, child: AddressableLeaf):
        if not isinstance(child, AddressableLeaf):
            raise TypeError('Cannot add a child that is not addressable')
        self.__children[child.getLocalAddress()] = child
        self.enabled = True

    def getLeafChildren(self) -> List[AddressableLeaf]:
        return self.getRecursiveChildren(leaveOnly=True)

    def getRecursiveChildren(self, leaveOnly=False) -> List[AddressableLeaf]:
        if not self.enabled:
            return []
        if self.isLeaf():
            return [self]

        children: List[AddressableLeaf] = []
        if not leaveOnly:
            children.append(self)
        for child in self.__children.values():
            if child.enabled:
                if isinstance(child, AddressableObject):
                    children.extend(child.getRecursiveChildren(leaveOnly=leaveOnly))
                elif isinstance(child, AddressableLeaf):
                    children.append(child)
        return children

    @property
    def children(self) -> List[AddressableLeaf]:
        return list(self.__children.values())

    def getByAddressString(self, addressString: str) -> Union[AddressableLeaf, bool]:
        if '/' not in addressString or addressString == '/':
            return super().getByAddressString(addressString)

        localAddress, _, childPath = addressString.partition('/')
        if not super().getByAddressString(localAddress):
            return False
        childAddress, _, _ = childPath.partition('/')
        if childAddress == '..':
            if self.parent is not None:
                return self.parent.getByAddressString(childPath)
            return self
        if childAddress in self.__children:
            return self.__children[childAddress].getByAddressString(childPath)
        return False

    def updateComplete(self) -> None:
        for child in self.__children.values():
            child.updateComplete()
        super().updateComplete()

    def saveToFile(self, filename: str) -> None:  # noqa: C901
        if filename.endswith(('.txt', '.TXT', '.text')):
            with open(filename, mode='w', encoding='utf8') as textfile:
                textfile.write(str(self))
        elif filename.endswith(('.htm', '.HTM', '.html', '.HTML')):
            with open(filename, mode='w', encoding='utf8'):
                raise ValueError('Object cannot be saved as HTML')
        elif filename.endswith(('.json')):
            with open(filename, mode='w', encoding='utf8') as textfile:
                textfile.write(self.toJSON() + '\n')
        elif filename.endswith(('.png', '.PNG')):
            with open(filename, mode='wb'):
                raise ValueError('Attribute is no image and cannot be converted to one')
        elif filename.endswith(('.jpg', '.JPG', '.jpeg', '.JPEG')):
            with open(filename, mode='wb'):
                raise ValueError('Attribute is no image and cannot be converted to one')
        else:
            raise ValueError('I cannot recognize the target file extension')


L = TypeVar('L', bound=AddressableLeaf)


class AddressableDict(AddressableObject, Dict[T, L]):
    def __setitem__(self, key: T, item: L):
        self.addChild(item)
        retVal = super().setdefault(key, item)
        if not self.enabled:
            self.enabled = True
        return retVal

    def __str__(self) -> str:
        return '[' + ', '.join([str(item) for item in self.values() if item.enabled]) + ']'


class AddressableList(AddressableObject, List[L]):
    def __add__(self, item):
        retVal = super().__add__(item)
        if not self.enabled:
            self.enabled = True
        return retVal

    def __str__(self) -> str:
        return '[' + ', '.join([str(item) for item in self if item.enabled]) + ']'



from datetime import datetime, time
import logging

from weconnect.addressable import AddressableAttribute, AddressableDict, AddressableObject
from weconnect.elements.enums import UnlockPlugState
from weconnect.elements.generic_settings import GenericSettings
from weconnect.elements.timer import Timer

LOG = logging.getLogger("weconnect")


class ChargingProfiles(GenericSettings):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.profiles = AddressableDict(localAddress='profiles', parent=self)
        self.timeInCar = AddressableAttribute(localAddress='timeInCar', parent=self, value=None, valueType=datetime)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update charging profiles from dict')

        if 'value' in fromDict:
            if 'profiles' in fromDict['value'] and fromDict['value']['profiles'] is not None:
                for profileDict in fromDict['value']['profiles']:
                    if 'id' in profileDict:
                        if profileDict['id'] in self.profiles:
                            self.profiles[profileDict['id']].update(fromDict=profileDict)
                        else:
                            self.profiles[profileDict['id']] = ChargingProfiles.ChargingProfile(fromDict=profileDict, parent=self.profiles)
                for profileId in [profileId for profileId in self.profiles.keys()
                                  if profileId not in [profile['id'] for profile in fromDict['value']['profiles'] if 'id' in profile]]:
                    del self.profiles[profileId]
            else:
                self.profiles.clear()
                self.profiles.enabled = False

            self.timeInCar.fromDict(fromDict['value'], 'timeInCar')
        else:
            self.profiles.clear()
            self.profiles.enabled = False
            self.timeInCar.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['profiles', 'timeInCar']))

    def __str__(self):
        string = super().__str__()
        if self.timeInCar.enabled:
            string += f'\n\tTime in Car: {self.timeInCar.value.isoformat()}'  # pylint: disable=no-member
            string += f' (captured at {self.carCapturedTimestamp.value.isoformat()})'  # pylint: disable=no-member
        string += f'\n\t\tProfiles: {len(self.profiles)} items'
        for profile in self.profiles.values():
            string += '\n' + ''.join(['\t\t\t' + line for line in str(profile).splitlines(True)])
        return string

    class ChargingProfile(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None
        ):
            self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=int)
            self.name = AddressableAttribute(localAddress='name', parent=self, value=None, valueType=str)
            self.maxChargingCurrent = AddressableAttribute(localAddress='maxChargingCurrent', parent=self, value=None, valueType=str)
            self.minSOC_pct = AddressableAttribute(localAddress='minSOC_pct', parent=self, value=None, valueType=int)
            self.targetSOC_pct = AddressableAttribute(localAddress='targetSOC_pct', parent=self, value=None, valueType=int)
            self.timers = AddressableDict(localAddress='timers', parent=self)
            self.preferredChargingTimes = AddressableDict(localAddress='preferredChargingTimes', parent=self)
            self.options = None
            super().__init__(localAddress=None, parent=parent)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):  # noqa: C901
            LOG.debug('Update charging profile from dict')

            if 'id' in fromDict:
                self.id.fromDict(fromDict, 'id')
                self.localAddress = str(self.id.value)
            else:
                LOG.error('Charging Proile is missing id attribute')

            self.name.fromDict(fromDict, 'name')
            self.maxChargingCurrent.fromDict(fromDict, 'maxChargingCurrent')
            self.minSOC_pct.fromDict(fromDict, 'minSOC_pct')
            self.targetSOC_pct.fromDict(fromDict, 'targetSOC_pct')

            if 'options' in fromDict and fromDict['options'] is not None:
                if self.options is not None and self.options.enabled:
                    self.options.update(fromDict=fromDict['options'])
                else:
                    self.options = ChargingProfiles.ChargingProfile.Options(fromDict=fromDict['options'], parent=self)

            if 'timers' in fromDict and fromDict['timers'] is not None:
                for chargingProfileTimerDict in fromDict['timers']:
                    if 'id' in chargingProfileTimerDict:
                        if chargingProfileTimerDict['id'] in self.timers:
                            self.timers[chargingProfileTimerDict['id']].update(fromDict=chargingProfileTimerDict)
                        else:
                            self.timers[chargingProfileTimerDict['id']] = Timer(
                                fromDict=chargingProfileTimerDict, parent=self.timers)
                for timerId in [timerId for timerId in self.timers.keys()
                                if timerId not in [timer['id']
                                for timer in fromDict['timers'] if 'id' in timer]]:
                    del self.timers[timerId]
            else:
                self.timers.clear()
                self.timers.enabled = False

            if 'preferredChargingTimes' in fromDict and fromDict['preferredChargingTimes'] is not None:
                for preferredChargingTimesDict in fromDict['preferredChargingTimes']:
                    if 'id' in preferredChargingTimesDict:
                        if preferredChargingTimesDict['id'] in self.preferredChargingTimes:
                            self.preferredChargingTimes[preferredChargingTimesDict['id']].update(fromDict=preferredChargingTimesDict)
                        else:
                            self.preferredChargingTimes[preferredChargingTimesDict['id']] = ChargingProfiles.ChargingProfile.PreferredTime(
                                fromDict=preferredChargingTimesDict, parent=self.preferredChargingTimes)
                for timeId in [timeId for timeId in self.preferredChargingTimes.keys()
                               if timeId not in [timer['id']
                               for timer in fromDict['timers'] if 'id' in timer]]:
                    del self.preferredChargingTimes[timeId]
            else:
                self.preferredChargingTimes.clear()
                self.preferredChargingTimes.enabled = False

            for key, value in {key: value for key, value in fromDict.items() if key not in ['id', 'name', 'maxChargingCurrent', 'minSOC_pct', 'targetSOC_pct',
                                                                                            'timers', 'preferredChargingTimes', 'options']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            string = ''
            if self.id.enabled:
                string += f'Profile: {self.id.value}'
            if self.name.enabled:
                string += f' - {self.name.value}'
            if self.options is not None and self.options.enabled:
                string += f'\nOptions: {self.options}'
            if self.timers.enabled:
                string += f'\nTimers: {len(self.timers)} items'
                for timer in self.timers.values():
                    string += ''.join(['\n\t' + line for line in str(timer).splitlines(True)])
            if self.preferredChargingTimes.enabled:
                string += f'\nPreferred Times: {len(self.preferredChargingTimes)} items'
                for preferredTime in self.preferredChargingTimes.values():
                    string += ''.join(['\n\t' + line for line in str(preferredTime).splitlines(True)])
            return string

        class PreferredTime(AddressableObject):
            def __init__(
                self,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress=None, parent=parent)
                self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=int)
                self.preferredTimeEnabled = AddressableAttribute(localAddress='enabled', parent=self, value=None, valueType=bool)
                self.startTime = AddressableAttribute(localAddress='startTime', parent=self, value=None, valueType=time)
                self.endTime = AddressableAttribute(localAddress='endTime', parent=self, value=None, valueType=time)
                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update preferred time from dict')

                if 'id' in fromDict:
                    self.id.fromDict(fromDict, 'id')
                    self.localAddress = str(self.id.value)
                else:
                    LOG.error('Preferred time is missing id attribute')

                self.preferredTimeEnabled.fromDict(fromDict, 'enabled')
                self.startTime.fromDict(fromDict, 'startTime')
                self.endTime.fromDict(fromDict, 'endTime')

                for key, value in {key: value for key, value in fromDict.items()
                                   if key not in ['id', 'enabled', 'startTime', 'endTime']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                string = ''
                if self.id.enabled:
                    string += f'{self.id.value}:'
                if self.preferredTimeEnabled.enabled:
                    string += f' Enabled: {self.preferredTimeEnabled.value}'
                if self.startTime.enabled:
                    string += f' Start: {self.startTime.value}'
                if self.endTime.enabled:
                    string += f' End: {self.endTime.value}'
                return string

        class Options(AddressableObject):
            def __init__(
                self,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress='options', parent=parent)
                self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=int)
                self.autoUnlockPlugWhenCharged = AddressableAttribute(localAddress='autoUnlockPlugWhenCharged', value=None, parent=self,
                                                                      valueType=UnlockPlugState)
                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update preferred time from dict')

                self.autoUnlockPlugWhenCharged.fromDict(fromDict, 'autoUnlockPlugWhenCharged')

                for key, value in {key: value for key, value in fromDict.items() if key not in ['autoUnlockPlugWhenCharged']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                string = ''
                if self.autoUnlockPlugWhenCharged.enabled:
                    string += f'\n\tAuto Unlock When Charged: {self.autoUnlockPlugWhenCharged.value.value}'
                return string



from weconnect.elements.control_operation import ControlInputEnum


class MaximumChargeCurrent(ControlInputEnum,):
    MAXIMUM = 'maximum'
    REDUCED = 'reduced'
    INVALID = 'invalid'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [MaximumChargeCurrent.MAXIMUM, MaximumChargeCurrent.REDUCED]


class UnlockPlugState(ControlInputEnum,):
    OFF = 'off'
    ON = 'on'
    PERMANENT = 'permanent'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [UnlockPlugState.OFF, UnlockPlugState.ON]



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()
INSTALL_REQUIRED = (HERE / "requirements.txt").read_text()
IMAGE_EXTRA_REQUIRED = (HERE / "image_extra_requirements.txt").read_text()
SETUP_REQUIRED = (HERE / "setup_requirements.txt").read_text()
TEST_REQUIRED = (HERE / "test_requirements.txt").read_text()

setup(
    name='weconnect',
    packages=find_packages(),
    version=open("weconnect/__version.py").readlines()[-1].split()[-1].strip("\"'"),
    description='Python API for the Volkswagen WeConnect Services',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Till Steinbach',
    keywords='weconnect, we connect, carnet, car net, volkswagen, vw, telemetry',
    url='https://github.com/tillsteinbach/WeConnect-python',
    project_urls={
        'Funding': 'https://github.com/sponsors/tillsteinbach',
        'Source': 'https://github.com/tillsteinbach/WeConnect-python',
        'Bug Tracker': 'https://github.com/tillsteinbach/WeConnect-python/issues'
    },
    license='MIT',
    install_requires=INSTALL_REQUIRED,
    extras_require={
        "Images": IMAGE_EXTRA_REQUIRED,
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='>=3.7',
    setup_requires=SETUP_REQUIRED,
    tests_require=TEST_REQUIRED,
    include_package_data=True,
    zip_safe=False,
)
